/*
 *  Add drag and zoom support for Eccharts for browsers that support pointer events.
 *
 *  Pointer events are not widely supported (Chrome 55, Firefox 59, MSIE10 - not safari)
 *  so we must include the polyfill pepjs for older browsers.
 *
 */
function Pinch(el, pinch, drag){

	this.callbacks = {
		pinch: pinch,
		drag: drag
	}

	el.addEventListener('pointerdown', this.bind(this.down))
	el.onpointermove = this.bind(this.move)
	el.onpointerup = this.bind(this.up)
	el.onpointercancel = this.bind(this.up)
	el.onpointerout = el.onpointerleave = this.bind(this.up)

	el.ontouchstart = el.ontouchend = el.ontouchmove = function(e){
		e.preventDefault();
		e.stopPropagation();
	}

	this.pointers = [];
	this.prevDiff = -1;
}

Pinch.prototype = {
	bind: function (fn) {
		var self = this;
		return function(){ return fn.apply(self, arguments); };
	},

	down: function(e){
		this.pointers.push({start: e, prev: e});
      e.preventDefault();
		e.stopPropagation();
	},

	up: function(e){
		this.remove_event(e);

		// If the number of pointers down is less than two then reset diff tracker
		if (this.pointers.length < 2){
			this.prevDiff = -1;

			if(this.op && this.op.type === 'pinch'){
				this.op = this.op.end()
			}

			if(this.pointers.length < 1){
				if(this.op && this.op.type === 'drag'){
					this.op = this.op.end(e)
				}
			}
		}

      e.preventDefault();
		e.stopPropagation();
	},

	move: function(e){
		var self = this;

		// Find this event in the cache and update its record with this event
		for (var p = 0; p < this.pointers.length; p++) {
			if (e.pointerId == this.pointers[p].prev.pointerId) {
				this.pointers[p].prev = e;
				break;
			}
		}

		function new_op(name){
			self.op = self.callbacks[name](e)
			self.op.type = name
		}

		switch(this.pointers.length){
			case 1:
				var start = this.pointers[0].start
				var diff = Math.abs(e.clientX - start.clientX) + Math.abs(this.pointers[0].start.clientY - e.clientY);
				console.log('pan', diff);
				if(diff > 20){
					this.op || new_op('drag')
					this.op.move(e) // TODO is right event? x,y already moved
				}
				break;

			case 2:
				// Two pointers are down, check for pinch gestures.

				// Calculate the distance between the two pointers
				var curDiff = Math.abs(this.pointers[0].prev.clientX - this.pointers[1].prev.clientX);

				if (this.prevDiff > 0) {
					var scale = this._scale(this.pointers[0], this.pointers[1])
					if(scale > 0.1 && scale < 10){
						this.op || new_op('pinch')
						this.op.move(scale)
					}
				}

				this.prevDiff = curDiff;
				break;
		}

      e.preventDefault();
		e.stopPropagation();
	},

	remove_event: function(e) {
		for (var i = 0; i < this.pointers.length; i++) {
			if (this.pointers[i].prev.pointerId == e.pointerId) {
				this.pointers.splice(i, 1);
				break;
			}
		}
	},

	_scale: function() {
		var p1 = this.pointers[0]
		var p2 = this.pointers[1]

		return this.distance(p1.prev, p2.prev) / this.distance(p1.start, p2.start);
	},

	distance: function(p1, p2) {
		var x = p2.clientX - p1.clientX;
		var y = p2.clientY - p1.clientY;

		return Math.sqrt((x * x) + (y * y));
	}
}

export default function PointerPlugin(chart, o) {
	var self = chart;

	setTimeout(function(){

		('ontouchstart' in document.documentElement) && console.log('_bind_image: ontouchstart YES');
		('ontouchstart' in document.documentElement) || console.log('_bind_image: ontouchstart NO');

		('onpointerdown' in document.documentElement) && console.log('_bind_image: onpointerdown YES');
		('onpointerdown' in document.documentElement) || console.warn('_bind_image: onpointerdown NO');

		new Pinch(
			chart.img[0],
			function(){
				console.log('   >> START PINCH <<')

				self.matrix.pinch_magics = self.matrix.magics;
				self.matrix.pinch_screen = self.matrix.screen;

				return {
					move: function(scale){
						// TODO only update on requestAnimationFrame
						self.pinch(scale);
					},
					end: function(){
					}
				}
			},
			function(e){
				self._mouseStart(e)

				return {
					move: function(e){
						self._mouseDrag(e)
					},
					end: function(e){
						self._mouseStop(e)
					}
				}
			}
		)
	}, 0)
};

