/*
 *  Chart widget using OpenLayers that provides compatibility
 *  with the original ECMWF chart widget, for example for sending
 *  and receiving bus messages. Compatibility may be removed
 *  in the future.
 */
import Bus from 'bus'
import Settings from 'settings'
import Product from 'chart/product'
import mixins from 'mixins'
import Observable from 'observable'
import Ajaxqueue from 'ajax_queue'

// http://openlayers.org/en/latest/apidoc/
import Map from 'ol/map'
import View from 'ol/view'
import Tile from 'ol/layer/tile'
import TileWMS from 'ol/source/tilewms'
import proj4 from 'proj4'
import Projection from 'ol/proj/projection'
import Proj from 'ol/proj';
import OLProj4 from 'ol/proj/proj4'
import Extent from 'ol/extent'
import TileGrid from 'ol/tilegrid/tilegrid'

import ImageLayer from 'ol/layer/image'
import ImageWMS from 'ol/source/imagewms'

var projection_name = 'EPSG:4326'
//var projection_name = 'EPSG:32661'
var TILING = true
var OLVERSION = 4

if(OLVERSION === 4){
	Proj.setProj4(proj4)
}
proj4.defs('EPSG:32661', '+proj=stere +lat_0=90 +lat_ts=90 +lon_0=0 +k=0.994 +x_0=2000000 +y_0=2000000 +datum=WGS84 +units=m +no_defs')
if(OLVERSION === 5){
	OLProj4.register(proj4)
}

if(projection_name == 'EPSG:4326'){
	var proj_extent = Proj.get(projection_name).getExtent();
	var start_resolution = Extent.getWidth(proj_extent) / 256;
	var resolutions = new Array(22);
	for (var i = 0, ii = resolutions.length; i < ii; ++i) {
		resolutions[i] = start_resolution / Math.pow(2, i);
	}
	var tile_grid = new TileGrid({
		extent: proj_extent, //[-180, -90, 180, 90],
		resolutions: resolutions,
		tileSize: [512, 512]
	})
}

var Chart = function(args) {
	var self = this;

	var o = this.options = extend({
		target: 'map',
		location: null, // type observable
		time: null, // type observable
		product: null,
		token: null,
		bus: Bus()
	}, args);

	this.numInFlightTiles = 0 // TODO this may be incorrect after removing layers

	o.bus
		.listen('layer-delete', function(_, msg) {
			self.remove_layer(msg.layer_id)
		})
		.listen('layer-options-change', function(_, msg) {
		})

	o.time.subscribe(this.bind(this.__set_time))

	this.map = new Map({
		target: o.target,
		view: new View({
			projection: projection_name,
			center: Proj.fromLonLat([0.0, 51.0], projection_name),
			zoom: 3
		})
	})

	this._ready = false;
	this.map.on('postrender', function(e) {
		if (!e.frameState) return

		var numHeldTiles = 0
		var wanted = e.frameState.wantedTiles
		for (var layer in wanted)
			if (wanted.hasOwnProperty(layer))
				numHeldTiles += Object.keys(wanted[layer]).length

		var ready = self.numInFlightTiles === 0 && numHeldTiles === 0

		if(ready && !self._ready) console.log('postrender: ready', e.map.tileQueue_)
		self._ready = ready
	})
}

Chart.prototype = extend({
	add_layer: function(layer) {
	},

	remove_layer: function(layer_id) {
		var layer = this.product.layers_by_id[layer_id];
		var index = this.product.layerList().indexOf(layer)

		if(index > -1){
			this._push_undo();
			this.product.layerDelete(layer_id);
			this._clear()._redraw(function() {
				this._notifyProductChange();
			});

			var l = this.map.getLayers().item(index)
			if(l) this.map.removeLayer(l)
		}
	},

	/*
	 *  Change to named product. Is very slow because an extra plot request is made to get the layer list.
	 */
	set_product: function(p) {
		var self = this, o = this.options;
		var old_layers = this.product ? this.product.layerList() : [];

		(this.product = new Product())
			.init({product: typeof p.name === 'function' ? p.name() : p.name}) // TODO enforce product object type
			.get_chart({
				json: {
					product: this.product.as_json({width: 128, height: 64}, this.time, this.mail, 'png')
				},
				success: function(product, time) {
					product = product.product
					var map_layers = self.map.getLayers()
					var new_layers = self.product.layerList()

					var vector = self.has_probe_layer() ? 1 : 0 // TODO preserve this layer

					// Truncate the layerlist if it is too long
					if(map_layers.getLength() > new_layers.length + vector){
						for(var i = map_layers.getLength() - 1; i >= new_layers.length; i--)
							map_layers.removeAt(i)
					}

					product.layers.forEach(function(v, i) {
						if((!old_layers[i]) || (v.name !== old_layers[i].name)){
							if(i < map_layers.getLength() - vector)
								map_layers.removeAt(i)

							if(TILING){
								var source = new TileWMS({
									url: Ajaxqueue.options.host + '/wms',
									params: {LAYERS: v.name, token: o.token},
									tileGrid: tile_grid,
								})

								source.on('tileloaderror', function() {
									console.error('tile failed')
									require(['notify'], function(notify) {
										notify({id: 1, kind: 'error', story: 'ok', message: 'WMS fetch failure'});
									})
								})
								source.on('tileloadstart', function () {++self.numInFlightTiles})
								source.on('tileloadend', function () {--self.numInFlightTiles})

								map_layers.insertAt(i, new Tile({
									source: source
								}))
							}else{
								map_layers.insertAt(i, new ImageLayer({
									source: new ImageWMS({
										url: Ajaxqueue.options.host + '/wms',
										params: {LAYERS: v.name, token: o.token}
									})
								}))
							}
						}
					})

					self.product.availability(time, function(availability) {
						o.time.set(time, availability.availability)
						o.bus.send('chart-product-availability', availability);
						o.bus.send('chart-content-update', {
							product: self.product,
							time: availability.current
						})
					})
				}
			})
	},

	set_projection: function(p) {
		// If using a projection whose coordinate unit is not lat/lon, the coords must be translated first, eg:
		// ol.proj.fromLonLat([x, y]);
		// ol.proj.transform([8.23, 46.86], 'EPSG:4326', 'EPSG:21781');

		var o = this.options

		// this regex will work only for some catalogue items
		var match = p.data.description.match(/ (\d+)S\/(\d+)W.* (\d+)N\/(\d+)E/)
		if(match){
			match.shift()
			var extent = [
				-match[1]
			]
			extent[1] = -match[0]
			extent[2] = +match[3]
			extent[3] = +match[2]

			// TODO get lat lon coordinate from api
			//var extent = [0, 40, 5, 50]
			this.map.getView().fit(extent, {size: this.map.getSize()})
		}else{
			console.log('TODO use a different regex');

			var match = p.data.description.match(/>(\d+.\d+)<\/span>\&deg;([NS]).*>(\d+.\d+)<\/span>\&deg;([WE]).*>(\d+.\d+)<\/span>\&deg;([NS]).*>(\d+.\d+)<\/span>\&deg;([EW])/)
			if(match){
				match.shift()
				// latlon: N is positive, S is negative, E is positive, W is negative
				var lat0 = +match[0]
				if(match[1] === 'S') lat0 = -lat0
				var lon0 = +match[2]
				if(match[3] === 'W') lon0 = -lon0
				var lat1 = +match[4]
				if(match[5] === 'S') lat1 = -lat1
				var lon1 = +match[6]
				if(match[7] === 'W') lon1 = -lon1

				var extent = [
					lat0,
					lon0,
					lat1,
					lon1,
				]

				this.map.getView().fit(extent, {size: this.map.getSize()})
			}
		}

		/*    This demonstrates how to use the new catalogue-items api
		 *    It does not support jsonp so we cannot use it yet.

		// list:
		/plots/plotapi/catalogue-items?token=
		{
			type: projection,
			user_items: true,
			standard_items: false
		}

		Ajaxqueue({
			url: '/plots/plotapi/catalogue-item?token=' + o.token,
			json: {
				name: p.data.name,
				type: 'projection',
			},
			success: function(result) {
			}
		})
		*/
	},

	__set_time: function(time) {
		var valid_time = time.valid_time.iso8601();
		var layers = this.product && this.product.layerList();
		this.map.getLayers().forEach(function(layer, i) {
			['TILE', 'IMAGE'].indexOf(layer.type) && layers[i] && layers[i].has_data && layer.getSource().updateParams({time: valid_time})
		})
	},

	has_probe_layer: function() {
		var vectors = this.map.getLayers().getArray().filter(function(layer){
			return layer.type === 'VECTOR'
		})
		vectors.length > 1 && console.warn('too many vector layers')
		return vectors.length === 1;
	},

	_push_undo: function() {
		return this
	},
	_clear: function() {
		return this
	},
	_redraw: function() {
		return this
	},
	_notifyProductChange: function() {
	},
}, mixins)

export default Chart;
