/*
 *  Widget to display a single meteogram
 *
 *  The meteogram can optionally have a 'parent' chart. If such a chart shares
 *  the same bus, it will cause the meteogram step to change when the
 *  'chart-content-update' message is sent.
 */
define(['config', 'jquery', 'ajax_queue', 'unit', 'bus', 'time', 'forecaster/meteograms', 'mixins', './meteogram.css', 'jquery-ui/draggable'], function(ECMWF, $, Ajaxqueue, Unit, Bus, Time, Meteograms, mixins) {

	var count = 0;

	$.widget('ui.meteogram', $.extend({

		options: {
			location: null, // type Observable
			time: null,
			bus: Bus(),
			priority: 'very-high',
			availability: false,
			product: '10d_epsgram_2t',
			width: null,
			height: 120,
			heightAdjustment: false,
			ready: $.noop,
			// Called when the image is loaded
			newAvailability: $.noop
		},

		_init: function () {
			var self = this, o = this.options;

			count++;

			self.name = 'meteogram_' + count;
			self.graphs = {};
			self._requested_size = {};
			self.time = o.time;
			self.is_location_changed = false
			self.ready = new $.Deferred() // this promise is not doing much at the moment but perhaps will add extra conditions?

			this.source = {} // the meteogram can optionally be associated with another plot (eg in Eccharts)

			this.element.addClass('meteogram').css({position: 'relative'});

			this.clip = this.element.div('.meteogram-clip')
				.css({position: 'absolute'})

			this.img = $(new Image())
				.addClass('meteogram-image')
				.css({position: 'absolute'})
				.appendTo(this.clip);

			self.redraw = $.throttle(500, function() {
				self.ready.then(self.bind(self._redraw))
			})

			o.location.subscribe(function(location) {
				if (self.element.is(':visible')) {
					self.is_location_changed = true;
					self.redraw();
				}
			})

			this.element.resize(function(e, size) {
				o.width = size.width || o.width;
				self.resize();
			});
			this.element.on('show', function() {
				self.resize();
			});

			console.log('meteogram: parent=', this.element.parent());

			self.resize = $.throttle(2000, function() {
				var viewport = self._viewport();
				if (viewport.width && (viewport.width !== self._requested_size.width)) {
					self.redraw();
					self._requested_size = viewport;
				}
			});

			if (o.product === 'vertical_profile') {
				this.check_date = this._check_date_vp
			}

			self.source.time = self.time;

			Meteograms.products().then(function(a) {
				var chart_product = Meteograms.product_by_name(o.product);

				self.active()

				self.family = chart_product.data.family.replace(/\s/g, '');
				if (!!~['efi', 'cdf'].indexOf(self.family)) {
					self.check_date = self._check_date_efi;
				}

				self.ready.resolve()
			});

			self.checkAvailability(o.availability);

			self.redraw()
		},

		setData: function(data) {
			var self = this;
			self.options = $.extend(self.options, data);
			self.redraw();
		},

		_redraw: function() {
			var self = this, o = this.options;
			var viewport = self._viewport();

			if (!self.element.is(':visible')) {
				return;
			}

			var json = $.extend({
					product: o.product,
					viewport: viewport,
					height_adjustment: o.heightAdjustment
				},
				viewport,
				o.location.value
			);

			// Depending on the meteogram type, sometimes magics respects the requested aspect ratio, sometimes not.
			json.viewport.height || (json.viewport.height = o.height);

			var is_product_available = true
			var is_new = !self.metadata

			function to_json(time){ // TODO this function already exists somewhere?
				time.base_time = $.dateToCycle(time.base_time);
				return time
			}

			if (self.time) { // The time comes from the plot, so the first request will never have self.time

				var time = {
					base_time: self.time.base_time,
					step: self.time.step
				}

				if (self.source.time) {

					is_product_available = self.check_date(time);
				}
				$.extend(json, to_json(time))
			}
			else console.log('meteogram: no self.time yet ...');

			if (json.viewport.height && json.viewport.width && (is_new || self.is_location_changed || is_product_available)) {
				Ajaxqueue({
					url: '/backend/meteogram/',
					json: json,
					status: ECMWF.ajax_loader(self.clip),
					priority: o.priority,
					logErrors: 'Meteogram',
					img: self.img,
					name: self.name,
					tag: self.name,
					success: function (json) {
						self.metadata = json.magics ? json.magics.metadata : {};
						console.log('meteogram: result=', json);
						self.availability = $.map(json.availability, function (x) {
							return new Time(x);
						});

						self.availability = self.availability.sort(function (a, b) {
							return a.base_time - b.base_time;
						});

						// only now that we have the availability can we in all cases show the correct meteogram
						if(!self.time) self.redraw()

						self.time = new Time(json.base_time, json.step, json.valid_time);

						//o.bus.send('meteogram:' + self.time.key);  // ---- ???

						self.element
							.stop()
							.animate({ height: self.img.height() }, o.ready);

						self._check_out_of_date();
						o.newAvailability(self.time, self.availability);

						if (json.title && o.title) {

							self.metadata.title = json.title;

							var adjusted = (self.metadata && self.metadata.temperature_correction)
								? ', adjusted to ' + self.metadata.height + 'm height'
								: ''

							setTimeout(function() {
								// TODO: something cleaner...
								var bt = $.dateToTitle(self.time.base_time);
								var vt = $.dateToTitle(self.time.valid_time);
								console.log('meteogram: metadata=', json.magics, json.magics.metadata.db);
								console.log('self.family: ', self.family);
								if (!!~['efi', 'cdf'].indexOf(self.family)) {
									console.warn('self.family in loop: ', self.family);

									var prev_day = self.time.valid_time.getTime() - 1000 * 60 * 60 * 25;

									prev_day = $.dateToTitle(new Date(prev_day));
									if (__DEV__)
										o.title.html(json.title + '<br>Base date: ' + bt + '<br>Valid for 24 hours from ' + prev_day + " to " + vt +  adjusted);
									else
										o.title.html(json.title + '<br>Base date: ' + bt + '<br>Valid for 24 hours from ' + prev_day + " to " + vt +  adjusted);
								}
								else {
									if (bt !== vt) o.title.html(json.title + '<br>Base date: ' + bt + ' Valid for: ' + vt + adjusted)
									else o.title.html(json.title + '<br>Base date: ' + bt + adjusted);
								}
							}, 10);
						}
						self.is_location_changed = false
					}
				})
			} else {
				console.log('meteogram not available for this step');
			}
		},

		showLatestCycle: function() {
			this.time = null;
			this.redraw();
		},

		/*
		_update_cycle: function(n) {
			var self = this;
			if (self.time && self.availability) {
				var idx = self.time.index(self.availability);
				idx += n;
				if (idx < 0) idx = 0;
				if (idx > self.availability.length - 1) idx = self.availability.length - 1;
				self.time = self.availability[idx];
				self.redraw();
			}
		},
		*/

		setHeightAdjustment: function(on) {
			var self = this, o = this.options;

			console.warn('setHeightAdjustment', on);
			o.heightAdjustment = on;
			self.redraw();
		},

		setCycle: function(time) {
			if (time) {
				this.time = time;
				this.redraw();
			}
		},

		previousStep: function() {
			var self = this, o = this.options;
			self.setCycle(self.time.previous(self.availability));
		},

		nextStep: function() {
			var self = this, o = this.options;
			self.setCycle(self.time.next(self.availability));
		},

		previousDay: function() {
			var self = this, o = this.options;

			if (self.time) {
				var prev_day = self.time.valid_time.getTime() - 1000 * 60 * 60 * 24;
				for(var prev = self.time; prev; prev = prev.previous(self.availability)){
					if(prev.valid_time.getTime() <= prev_day){
						self.setCycle(prev);
						break;
					}
				}
			}
		},

		nextDay: function() {
			var self = this, o = this.options;

			if (self.time) {
				var next_day = self.time.valid_time.getTime() + 1000 * 60 * 60 * 24;
				for(var next = self.time; next; next = next.next(self.availability)){
					if(next.valid_time.getTime() >= next_day){
						self.setCycle(next);
						break;
					}
				}
			}
		},

		previousCycle: function() {
			var self = this, o = this.options;

			console.log('previousCycle');
			self.setCycle(self.time.previousBaseTime(self.availability, true));
		},

		nextCycle: function() {
			var self = this, o = this.options;

			console.log('nextCycle');
			self.setCycle(self.time.nextBaseTime(self.availability, true));
		},

		_check_out_of_date: function() {
			var self = this, o = this.options;

			if (self.time && self.availability) {
				var latest_basetime = self.availability[self.availability.length - 1].base_time;
				var selected_basetime = self.time.base_time;
				o.bus.send('meteogram-out-of-date', selected_basetime - latest_basetime !== 0);
			}
		},

		checkAvailability: function(on) {
			var self = this, o = this.options;

			if (on) {
				if (!self.timer) {

					self.timer = setInterval(function () {

						if (!self.element.is(':visible')) return;

						Ajaxqueue({
							url: '/backend/meteogram-availability/',
							json: {
								product: o.product
							},
							priority: 'low',
							logErrors: 'Meteogram availability',
							success: function(json) {
								if (json) {
									var avail = $.map(json.availability, function(t) {
										return new Time(t);
									});
									self.availability = avail;

									self.availability = self.availability.sort(function(a, b) {
										return a.base_time - b.base_time;
									});

									o.newAvailability(self.time, self.availability);
									self._check_out_of_date();
								}
							}
						});
					},
					60 * 1000);
				}
			}
			else {
				if (self.timer) clearInterval(self.timer);
				self.timer = null;
			}
		},

		check_date: function(t) {
			var self = this;
			var is_product_available = true;

			if (this.time.base_time.getTime() !== this.source.time.base_time.getTime()) {
				// base time has changed

				var time = this.source.time || this.time
				var best_time = self.find_best_time(new Time(time.base_time, 0, time.base_time));
				if (best_time) {
					t.step = best_time.step;
					t.base_time = best_time.base_time;
				}
			}
			else
				is_product_available = false;

			return is_product_available;
		},

		_check_date_vp: function(t) {
			var best_time = this.find_best_time(this.source.time);

			if (best_time) {
				t.base_time = best_time.base_time;
				t.step = best_time.step;

				$('#na').remove()

				return true
			} else {
				t.base_time = this.time.base_time;
				t.step = this.time.step;

				this.availability && this.clip.prepend(
					$('#na')
						.orelse(function(){ return $('<div>').attr({id: 'na'}) })
						//.html('Vertical profile not available for valid time ' + $.dateToTitle(this.source.time.valid_time) + ' or location')
						.html('Vertical profile not available for selected valid time or location')
				)

				// Ignore location changes for if the valid time is not available
				// This can perhaps be improved later by updating the location with the closest valid time.
				this.is_location_changed = false;

				return false
			}
		},

		_check_date_efi: function(t) {
			var is_product_available = true;
			t.base_time = this.time.base_time;
			t.step = this.source.time.step;
			var stp = this.source.time.step;
			var addition = 0;
			if (this.source.time.base_time.getHours() === 12) {
				//console.log('hour=12');
				if (stp < 12) stp = 0
				else stp = stp - 12
					addition = 12;
			}
			//var step = Math.ceil((stp - 1) / 24.) * 24;
			var step = parseInt(Math.ceil((stp - 1) / 24.)) * 24 + addition;
			//console.warn('STEP: ', step);

			if(step > 120){
				console.log('WE SHOULD NOT MAKE REQUEST');
				this.is_location_changed = false;
				return false
			}

			var vt = new Date(this.source.time.base_time.getTime() + (step-1)*60*60*1000);
			var newtime = {
				key     : this.source.time.base_time.yyyymmddHHMM() + '_' + $.dateToCycle(vt),
				step    : step,
				base_time   : this.source.time.base_time,
				valid_time  : vt
			}
			//console.log('meteogram.efi: NEWTIME:', newtime);
			var best_time = this.find_best_time(newtime);
			//console.log('meteogram.efi: BESTTIME:', best_time);
			if (best_time) {
				t.step = best_time.step;
				t.base_time = best_time.base_time;

				is_product_available = true;
			}
			else {
				$.extend(t, newtime);
				is_product_available = true;
			}
			return is_product_available;
		},

		active: function(active) {
			var self = this, o = this.options, bus = Bus();

			if(active !== false){
				bus
					// forgetting then rebinding means that we will get a PAST message giving the current chart state
					.forget_me(self, 'chart-content-update')
					.listen(self, 'chart-content-update', function(event, msg) {
						self.source.time = msg.time;// || self.time;
						self.redraw();
					})
			}else{
				bus.forget_me(self, 'chart-content-update');
			}
		},

		getHelp: function() {
			var self = this, o = this.options;
			var list = [];

			if (!self.metadata) {
				return 'Cannot get help information.';
			}

			if (!self.metadata.distance) {
				return 'Cannot display help information. Missing Distance data.';
			}

			var m = Unit.get('m');
			var km = Unit.get(self.metadata.distance.distance_unit || 'km');

			var gridpoint_type = (self.metadata.land_sea_mask < 0.5) ? 'sea' : 'land';

			//New format for Meteogram metadata as in WebAPI
			/*
			location =>  ens_location
			eps_height => ens_height
			deterministic_height => hres_height
			*/

			list.push("### " + self.metadata.title );
			list.push("Location requested: **" + Unit.formatGeoLocation(self.metadata.requested_location) + "**, ");
			list.push("nearest EPS grid point: **" + Unit.formatGeoLocation(self.metadata.location || self.metadata.ens_location) + '**.');
			list.push('');
			list.push("Height of requested location (based on the GTOPO30 dataset): **" + m.asHTML(self.metadata.height) + "**, ");
			list.push("height of EPS orography at nearest grid point: **" + m.asHTML(self.metadata.eps_height || self.metadata.ens_height) + "**, ");
			list.push("height of deterministic orography interpolated to EPS resolution at nearest grid point: **" + m.asHTML(self.metadata.deterministic_height ||
			self.metadata.hres_height) + '**.');
			if (self.metadata.temperature_correction) {
				list.push('');
				if(self.metadata.temperature_adjustment) {
					list.push('Temperature parameter is adjusted to the requested location by **' + $.formatCelcius(self.metadata.temperature_adjustment.deterministic_adjustement, 1) + '**' + ' in deterministic forecast curve and ' + '**' + $.formatCelcius(self.metadata.temperature_adjustment.eps_adjustement, 1) + '** in EPS forecast curves.');
				}
				else
				{
					list.push('Temperature parameter is adjusted to the requested location');
				}
				list.push('');
			}

			list.push('');
			list.push('');
			//TODO: Fix me later
			if (self.metadata.db === "10_days_wave") {
				list.push("The nearest EPS grid point is located at **" + km.asHTML(self.metadata.distance.distance) + " " + self.metadata.distance.direction + "** of the selected location.");
			}
			else {
				list.push("The nearest EPS grid point is a **" + gridpoint_type + " point**, ");
				list.push("located at **" + km.asHTML(self.metadata.distance.distance) + " " + self.metadata.distance.direction + "** of the selected location.");
			}

			/*
			list.push('TODO:');
			list.push('');
			list.push('* Land/point or sea point (DONE)');
			list.push('* Distance to point (DONE)');
			list.push('* Model resolutions (Monday the 28th)');
			list.push('* Explain red/blue curves(Monday the 28th)');
			list.push('');
			*/

			/* $.each(self.metadata, function(k,v) { list.push("* " + k + " = " + v); }); */

			return list.join("\n");
		},

		_destroy: function() {
			if (this.timer) clearInterval(this.timer);

			this.element.removeClass('meteogram');
			this.clip.remove();
		},

		_viewport: function() {
			return {
				top: 0,
				left: 0,
				width: this.options.width || this.element.width(),
				height: this.element.height()
			};
		},

		/*
		 *  Find best time, same valid time, latest bt
		 */
		find_best_time: function(time) {
			var availability = this.availability;
			if(__DEV__ && !availability) console.warn('find_best_time: no availability');
			var best_time;
			if(this.availability){

				for (var i in availability) {
					if (time.key === availability[i].key)
						return availability[i];
				}

				var valid_time = time.valid_time;
				for (i in availability) {
					if (valid_time.getTime() === availability[i].valid_time.getTime()) {
						if (! best_time)
							best_time = availability[i];
						if ($.dateToCycle(availability[i].base_time) > $.dateToCycle(best_time.base_time))
							best_time = availability[i];
					}
				}
			}
			return best_time;
		}
	},
	{bind: mixins.bind}));
});
